/*
 * Copyright (C) 2025  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/check_value.hh"

#include <algorithm>
#include <regex>

namespace {

constexpr bool is_utf8_leading_byte(char byte)
{
    return (byte & 0xc0) != 0x80;
}

static_assert(is_utf8_leading_byte(" "[0]), "must be leading byte");
static_assert(is_utf8_leading_byte("a"[0]), "must be leading byte");
static_assert(is_utf8_leading_byte("á"[0]), "must be leading byte");
static_assert(!is_utf8_leading_byte("á"[1]), "cannot be leading byte");

//normalizedString: A string that does not contain line feeds, carriage returns, or tabs
template <typename S, typename L>
bool is_normalized_string(const std::string& value, S is_too_short, L is_too_long)
{
    std::size_t length = 0;
    if (is_too_long(length))
    {
        return true;
    }
    const bool is_ok = std::none_of(begin(value), end(value), [&length, &is_too_long](char c)
    {
        if (is_utf8_leading_byte(c))
        {
            ++length;
            if (is_too_long(length))
            {
                return true;
            }
            switch (c)
            {
                case '\r': // carriage return
                case '\n': // line feed
                case '\t': // horizontal tab
                case '\v': // vertical tab
                    return true;
            }
        }
        return false;
    });
    return is_ok && !is_too_short(length);
}

//token: A string that does not contain line feeds, carriage returns, tabs, leading or trailing spaces, or multiple spaces
template <typename S, typename L>
bool is_token(const std::string& value, S is_too_short, L is_too_long)
{
    if (value.empty())
    {
        return !is_too_short(0);
    }
    if ((value.front() == ' ') ||
        (value.back() == ' '))
    {
        return false;
    }
    std::size_t length = 0;
    if (is_too_long(length))
    {
        return true;
    }
    bool previous_is_space = false;
    const bool is_ok = std::none_of(
            begin(value),
            end(value),
            [&length, &is_too_long, &previous_is_space](char c)
    {
        if (is_utf8_leading_byte(c))
        {
            ++length;
            if (is_too_long(length))
            {
                return true;
            }
            switch (c)
            {
                case '\r': // carriage return
                case '\n': // line feed
                case '\t': // horizontal tab
                case '\v': // vertical tab
                    return true;
                case ' ':
                {
                    const bool two_spaces_in_row = previous_is_space;
                    previous_is_space = true;
                    return two_spaces_in_row;
                }
                default:
                    previous_is_space = false;
            }
        }
        return false;
    });
    return is_ok && !is_too_short(length);
}

}//namespace {anonymous}

using namespace Fred::Registry;

bool Fred::Registry::is_country_code_valid(const LibFred::OperationContext& ctx, const CountryCode& cc)
{
    return 0 < ctx.get_conn().exec_params(
                "SELECT FROM enum_country WHERE id = $1::TEXT FOR SHARE",
                Database::query_param_list(get_raw_value_from(cc))).size();
}

//<simpleType name="postalLineType">
//  <restriction base="normalizedString">
//    <minLength value="1"/>
//    <maxLength value="255"/>
//  </restriction>
//</simpleType>
bool Fred::Registry::is_postal_line_type(const std::string& value)
{
    return is_normalized_string(
            value,
            [](auto&& length) { return length < 1; },
            [](auto&& length) { return 255 < length; });
}

//<simpleType name="optPostalLineType">
//  <restriction base="normalizedString">
//    <maxLength value="255"/>
//  </restriction>
//</simpleType>
bool Fred::Registry::is_opt_postal_line_type(const std::string& value)
{
    return is_normalized_string(
            value,
            [](auto&&) { return false; },
            [](auto&& length) { return 255 < length; });
}

//<simpleType name="pcType">
//  <restriction base="token">
//    <maxLength value="16"/>
//  </restriction>
//</simpleType>
bool Fred::Registry::is_pc_type(const Fred::Registry::PostalCode& value)
{
    return is_token(
            get_raw_value_from(value),
            [](auto&&) { return false; },
            [](auto&& length) { return 16 < length; });
}

//<simpleType name="e164StringType">
//  <restriction base="token">
//    <pattern value="(\+[0-9]{1,3}\.[0-9]{1,14})?"/>
//    <maxLength value="17"/>
//  </restriction>
//</simpleType>
bool Fred::Registry::is_e164_string_type(const Fred::Registry::PhoneNumber& phone_number)
{
    static const auto regex = std::regex{"\\+[0-9]{1,3}\\.[0-9]{1,14}"};
    const auto& value = get_raw_value_from(phone_number);
    std::smatch m;
    return is_token(
            value,
            [](auto&&) { return false; },
            [](auto&& length) { return 17 < length; }) &&
           std::regex_match(value, m, regex);
}

//<simpleType name="emailCommaListType">
//  <restriction base="token">
//    <pattern value="[^@, ]{1,64}@[^@, ]+(,[^@, ]{1,64}@[^@, ]+)*"/>
//    <maxLength value="320"/>
//  </restriction>
//</simpleType>
bool Fred::Registry::is_email_comma_list_type(const std::vector<Fred::Registry::EmailAddress>& values)
{
    static const auto regex = std::regex{"[^@, ]{1,64}@[^@, ]+"};
    std::size_t total_length = 0;
    return std::all_of(begin(values), end(values), [&total_length](auto&& value)
    {
        const std::string& email = get_raw_value_from(value);
        std::size_t email_length = 0;
        if (!is_token(
                email,
                [&email_length](auto&& length)
                {
                    email_length = length;
                    return length < 3;
                },
                [&total_length](auto&& length) { return 320 < (total_length + length); }))
        {
            return false;
        }
        if (total_length == 0)
        {
            total_length = email_length;
        }
        else
        {
            total_length += 1 + email_length;
            if (320 < total_length)
            {
                return false;
            }
        }
        std::smatch m;
        return std::regex_match(email, m, regex);
    });
}

//<simpleType name="vatT">
//  <restriction base="token">
//    <maxLength value="20"/>
//  </restriction>
//</simpleType>
bool Fred::Registry::is_vat_t(const Fred::Registry::VatIdentificationNumber& value)
{
    return is_token(
            get_raw_value_from(value),
            [](auto&&) { return false; },
            [](auto&& length) { return 20 < length; });
}

//<simpleType name="identValueT">
//  <restriction base="token">
//    <maxLength value="32"/>
//  </restriction>
//</simpleType>
bool Fred::Registry::is_ident_value_t(const std::string& value)
{
    return is_token(
            value,
            [](auto&&) { return false; },
            [](auto&& length) { return 32 < length; });
}
