/*
 * Copyright (C) 2018-2025  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/contact/contact_state.hh"

#include "src/fred_unwrap.hh"
#include "src/fred_wrap.hh"
#include "src/util/strong_type.hh"
#include "src/util/struct_to_string.hh"

#include "libfred/db_settings.hh"
#include "libfred/registrable_object/contact/get_contact_state.hh"
#include "libfred/registrable_object/snapshot_id.hh"
#include "util/db/db_exceptions.hh"

#include <boost/uuid/string_generator.hpp>

#include <utility>

namespace Fred {
namespace Registry {
namespace Contact {

ContactId ContactStateRequest::get_contact_id() const
{
    try
    {
        return Util::make_strong<ContactId>(boost::uuids::string_generator{}(contact_id));
    }
    catch (...)
    {
        throw ContactStateReply::Exception::InvalidData{*this, {"contact_id"}};
    }
}

std::string ContactStateRequest::to_string() const
{
    return Util::StructToString().add("contact_id", contact_id)
                                 .add("contact_snapshot_id", contact_snapshot_id)
                                 .finish();
}

std::string ContactState::to_string() const
{
    if (flag_presents.empty())
    {
        return "{ }";
    }
    std::string buffer;
    for (const auto& presents : flag_presents)
    {
        if (!buffer.empty())
        {
            buffer += ", ";
        }
        buffer += "'" + presents.first + "': " + (presents.second ? "presents" : "absents");
    }
    return "{ " + buffer + " }";
}

std::string ContactStateReply::Data::to_string() const
{
    return Util::StructToString().add("contact_id", contact_id)
                                 .add("state", state)
                                 .add("snapshot_id", snapshot_id)
                                 .finish();
}

ContactStateReply::Exception::ContactDoesNotExist::ContactDoesNotExist(const ContactStateRequest& src)
    : Registry::Exception("contact " + src.to_string() + " does not exist")
{
}

ContactStateReply::Exception::InvalidSnapshot::InvalidSnapshot(const ContactStateRequest& src)
    : Registry::Exception{"snapshot_id of contact " + Util::Into<std::string>::from(src.contact_id) + " "
                          "differs from current version"}
{ }

ContactStateReply::Exception::InvalidData::InvalidData(const ContactStateRequest& src, std::set<std::string> fields)
    : Registry::Exception{"invalid data: " + src.to_string()},
      fields{std::move(fields)}
{ }

ContactStateReply::Data contact_state(const ContactStateRequest& request)
{
    try
    {
        LibFred::OperationContextCreator ctx;
        LibFred::OperationContextLockingForShare ctx_for_share{ctx};
        ContactStateReply::Data result;
        const auto contact_id = request.get_contact_id();
        const auto contact_uuid = fred_wrap(contact_id);
        if (request.contact_snapshot_id != boost::none)
        {
            try
            {
                using CheckObjectChange = LibFred::RegistrableObject::CheckObjectChange;
                const auto check_result = CheckObjectChange{contact_uuid}
                    .exec(ctx_for_share, fred_wrap(*request.contact_snapshot_id));
                if (check_result.data_has_been_changed || check_result.state_has_been_changed)
                {
                    throw ContactStateReply::Exception::InvalidSnapshot{request};
                }
            }
            catch (const LibFred::RegistrableObject::InvalidSnapshotId&)
            {
                throw ContactStateReply::Exception::InvalidData{request, {"contact_snapshot_id"}};
            }
            catch (const LibFred::UnknownObjectId&)
            {
                throw ContactStateReply::Exception::ContactDoesNotExist{request};
            }
        }
        result.state = fred_unwrap(LibFred::RegistrableObject::Contact::GetContactStateByUuid(contact_uuid).exec(ctx_for_share));
        result.contact_id = contact_id;
        result.snapshot_id = fred_unwrap(LibFred::RegistrableObject::get_snapshot_id(ctx));
        ctx.commit_transaction();
        return result;
    }
    catch (const LibFred::RegistrableObject::Contact::GetContactStateByUuid::DoesNotExist&)
    {
        throw ContactStateReply::Exception::ContactDoesNotExist(request);
    }
    catch (const ContactStateReply::Exception::ContactDoesNotExist&)
    {
        throw;
    }
    catch (const ContactStateReply::Exception::InvalidData&)
    {
        throw;
    }
    catch (const ContactStateReply::Exception::InvalidSnapshot&)
    {
        throw;
    }
    catch (const Database::Exception& e)
    {
        throw Registry::DatabaseError(std::string("Database::Exception: ") + e.what());
    }
    catch (const std::exception& e)
    {
        throw Registry::InternalServerError(std::string("expected unknown exception: ") + e.what());
    }
    catch (...)
    {
        throw Registry::InternalServerError("unexpected unknown exception");
    }
}

}//namespace Fred::Registry::Contact
}//namespace Fred::Registry
}//namespace Fred
