/*
 * Copyright (C) 2020-2025  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

 #include "src/contact/update_contact.hh"

#include "src/check_value.hh"
#include "src/fred_unwrap.hh"
#include "src/fred_wrap.hh"
#include "src/util/into.hh"
#include "src/util/struct_to_string.hh"

#include "libfred/contact_verification/django_email_format.hh"
#include "libfred/notifier/enqueue_notification.hh"
#include "libfred/object_state/cancel_object_state_request_id.hh"
#include "libfred/object_state/lock_object_state_request_lock.hh"
#include "libfred/object_state/perform_object_state_request.hh"
#include "libfred/opcontext.hh"
#include "libfred/poll/create_poll_message.hh"
#include "libfred/registrable_object/contact/get_contact_state.hh"
#include "libfred/registrable_object/contact/info_contact.hh"
#include "libfred/registrable_object/contact/update_contact.hh"
#include "libfred/registrable_object/snapshot_id.hh"
#include "libfred/registrar/info_registrar.hh"
#include "util/db/db_exceptions.hh"
#include "util/log/log.hh"

#include <boost/algorithm/algorithm.hpp>

#include <algorithm>
#include <utility>

namespace Fred {
namespace Registry {
namespace Contact {

std::string SetEmailAddresses::to_string() const
{
    return Util::StructToString().add("values", values)
        .finish();
}

ContactId UpdateContactRequest::get_contact_id() const
{
    try
    {
        return Util::make_strong<ContactId>(boost::uuids::string_generator{}(contact_id));
    }
    catch (...)
    {
        throw UpdateContactReply::Exception::InvalidData{*this, {"contact_id"}};
    }
}

std::string UpdateContactRequest::to_string() const
{
    return Util::StructToString().add("contact_id", contact_id)
                                 .add("contact_snapshot_id", contact_snapshot_id)
                                 .add("set_name", set_name)
                                 .add("set_name_privacy", set_name_privacy)
                                 .add("set_organization", set_organization)
                                 .add("set_organization_privacy", set_organization_privacy)
                                 .add("set_place", set_place)
                                 .add("set_place_privacy", set_place_privacy)
                                 .add("set_telephone", set_telephone)
                                 .add("set_telephone_privacy", set_telephone_privacy)
                                 .add("set_fax", set_fax)
                                 .add("set_fax_privacy", set_fax_privacy)
                                 .add("set_emails", set_emails)
                                 .add("set_emails_privacy", set_emails_privacy)
                                 .add("set_notify_emails", set_notify_emails)
                                 .add("set_notify_emails_privacy", set_notify_emails_privacy)
                                 .add("set_vat_identification_number", set_vat_identification_number)
                                 .add("set_vat_identification_number_privacy", set_vat_identification_number_privacy)
                                 .add("set_additional_identifier", set_additional_identifier)
                                 .add("set_additional_identifier_privacy", set_additional_identifier_privacy)
                                 .add("set_mailing_address", set_mailing_address)
                                 .add("set_billing_address", set_billing_address)
                                 .add("set_shipping_address1", set_shipping_address[0])
                                 .add("set_shipping_address2", set_shipping_address[1])
                                 .add("set_shipping_address3", set_shipping_address[2])
                                 .add("set_warning_letter", set_warning_letter)
                                 .add("log_entry_id", log_entry_id)
                                 .add("by_registrar", by_registrar)
                                 .finish();
}

std::string UpdateContactReply::Data::to_string() const
{
    return Util::StructToString().add("contact_handle", contact_handle)
                                 .add("contact_id", contact_id)
                                 .add("contact_history_id", contact_history_id)
                                 .add("events", events)
                                 .finish();
}

UpdateContactReply::Exception::ContactDoesNotExist::ContactDoesNotExist(const UpdateContactRequest& src)
    : Registry::Exception("contact " + Util::Into<std::string>::from(src.contact_id) + " does not exist")
{
}

UpdateContactReply::Exception::RegistrarDoesNotExist::RegistrarDoesNotExist(const UpdateContactRequest& src)
    : Registry::Exception("registrar " + Util::Into<std::string>::from(src.by_registrar) + " does not exist")
{
}

UpdateContactReply::Exception::InvalidSnapshot::InvalidSnapshot(const UpdateContactRequest& src)
    : Registry::Exception("snapshot_id of contact " + Util::Into<std::string>::from(src.contact_id) + " "
                          "differs from current version")
{
}

UpdateContactReply::Exception::UnauthorizedRegistrar::UnauthorizedRegistrar(const UpdateContactRequest&)
    : Registry::Exception("unauthorized registrar")
{
}

UpdateContactReply::Exception::UpdateProhibited::UpdateProhibited(const UpdateContactRequest&)
    : Registry::Exception("update prohibited")
{
}

UpdateContactReply::Exception::InvalidData::InvalidData(const UpdateContactRequest& src, std::set<std::string> fields)
    : Registry::Exception{"invalid data: " + src.to_string()},
      fields{std::move(fields)}
{ }

namespace {

template <typename T, typename OperationSet>
void optional_set(
        const boost::optional<T>& operation,
        OperationSet&& on_set)
{
    if (operation != boost::none)
    {
        std::forward<OperationSet>(on_set)(*operation);
    }
}

struct InvalidEmail : std::exception
{
    const char* what() const noexcept override { return "invalid email format"; }
};

struct NoEmail : std::exception
{
    const char* what() const noexcept override { return "no email"; }
};

decltype(auto) join_emails(const std::vector<EmailAddress>& addresses)
{
    std::string result;
    static const auto get_valid_email = [](const EmailAddress& email)
    {
        const auto& raw_email = get_raw_value_from(email);
        static DjangoEmailFormat django_email_format{};//constructor is not constexpr and method check is not const!
        const bool email_is_valid = django_email_format.check(raw_email);
        if (!email_is_valid)
        {
            FREDLOG_INFO("email \"" + raw_email + "\" is not valid");
            throw InvalidEmail{};
        }
        return raw_email;
    };
    for (const auto& email : addresses)
    {
        if (result.empty())
        {
            result = get_valid_email(email);
        }
        else
        {
            result += "," + get_valid_email(email);
        }
    }
    return result;
}

decltype(auto) to_libfred_contact_address(const ContactAddress& src)
{
    LibFred::ContactAddress dst;
    if (!src.company.empty())
    {
        dst.company_name = src.company;
    }
    if (0 < src.street.size())
    {
        dst.street1 = src.street[0];
        if (1 < src.street.size())
        {
            dst.street2 = src.street[1];
            if (2 < src.street.size())
            {
                dst.street3 = src.street[2];
            }
        }
    }
    dst.city = src.city;
    if (!src.state_or_province.empty())
    {
        dst.stateorprovince = src.state_or_province;
    }
    dst.postalcode = get_raw_value_from(src.postal_code);
    dst.country = get_raw_value_from(src.country_code);
    return dst;
}

bool is_country_code_valid(const LibFred::OperationContext& ctx, const std::string& cc)
{
    return 0 < ctx.get_conn().exec_params("SELECT FROM enum_country WHERE id = $1::TEXT", Database::QueryParams{cc}).size();
}

template <typename Fnc>
void check_country_code(const LibFred::OperationContext& ctx, const CountryCode& country_code, Fnc&& on_error)
{
    if (!is_country_code_valid(ctx, get_raw_value_from(country_code)))
    {
        std::forward<Fnc>(on_error)(country_code);
    }
}

template <typename A, typename Fnc>
void check_country_code(const LibFred::OperationContext& ctx, const A& address, Fnc&& on_error)
{
    check_country_code(ctx, address.country_code, std::forward<Fnc>(on_error));
}

bool is_empty(const std::string& value)
{
    return value.empty();
}

template <typename T>
bool is_empty(const std::vector<T>& array)
{
    return array.empty() ||
           std::all_of(begin(array), end(array), [](auto&& item) { return is_empty(item); });
}

template <typename N, template <typename> class ...Skills>
bool is_empty(const Util::StrongType<std::string, N, Skills...>& value)
{
    return is_empty(get_raw_value_from(value));
}

bool is_empty(const ContactAddress& value)
{
    return is_empty(value.company) &&
           is_empty(value.street) &&
           is_empty(value.city) &&
           is_empty(value.state_or_province) &&
           is_empty(value.postal_code) &&
           is_empty(value.country_code);
}

bool is_empty(const ContactAdditionalIdentifier& value)
{
    return boost::apply_visitor([](auto& v) { return is_empty(v); }, value);
}

void trim(std::string& value)
{
    boost::trim(value);
}

template <typename N, template <typename> class ...Skills>
void trim(Util::StrongType<std::string, N, Skills...>& value)
{
    value = Util::make_strong<Util::StrongType<std::string, N, Skills...>>(boost::trim_copy(get_raw_value_from(value)));
}

template <typename T>
void trim(std::vector<T>& array)
{
    std::for_each(begin(array), end(array), [](auto& value) { trim(value); });
}

void trim(SetEmailAddresses& set)
{
    trim(set.values);
}

void trim(PlaceAddress& value)
{
    trim(value.street);
    trim(value.city);
    trim(value.state_or_province);
    trim(value.postal_code);
    trim(value.country_code);
}

void trim(ContactAddress& value)
{
    trim(value.company);
    trim(value.street);
    trim(value.city);
    trim(value.state_or_province);
    trim(value.postal_code);
    trim(value.country_code);
}

auto trim_copy(ContactAddress value)
{
    trim(value);
    return value;
}

auto trim_copy(ContactAdditionalIdentifier value)
{
    boost::apply_visitor([](auto& v) { trim(v); }, value);
    return value;
}

template <typename T>
auto trim_copy(boost::optional<T> value)
{
    if (value != boost::none)
    {
        trim(*value);
    }
    return value;
}

template <typename T>
auto trim_copy(boost::optional<Util::Deletable<T>> value)
{
    if (value != boost::none &&
        get_raw_value_from(*value) != boost::none)
    {
        auto trimmed_copy = trim_copy(*get_raw_value_from(*value));
        if (!is_empty(trimmed_copy))
        {
            value = Util::to_set(std::move(trimmed_copy));
        }
        else
        {
            value = Util::to_delete<T>();
        }
    }
    return value;
}

template <typename T, std::size_t n>
auto trim_copy(std::array<boost::optional<Util::Deletable<T>>, n> array)
{
    std::for_each(begin(array), end(array), [](auto& value)
    {
        if (value != boost::none &&
            get_raw_value_from(*value) != boost::none)
        {
            auto trimmed_copy = trim_copy(*get_raw_value_from(*value));
            if (!is_empty(trimmed_copy))
            {
                value = Util::to_set(std::move(trimmed_copy));
            }
            else
            {
                value = Util::to_delete<T>();
            }
        }
    });
    return array;
}

struct TrimmedChange
{
    explicit TrimmedChange(const UpdateContactRequest& request)
        : set_name{trim_copy(request.set_name)},
          set_organization{trim_copy(request.set_organization)},
          set_place{trim_copy(request.set_place)},
          set_telephone{trim_copy(request.set_telephone)},
          set_fax{trim_copy(request.set_fax)},
          set_emails{trim_copy(request.set_emails)},
          set_notify_emails{trim_copy(request.set_notify_emails)},
          set_vat_identification_number{trim_copy(request.set_vat_identification_number)},
          set_additional_identifier{trim_copy(request.set_additional_identifier)},
          set_mailing_address{trim_copy(request.set_mailing_address)},
          set_billing_address{trim_copy(request.set_billing_address)},
          set_shipping_address{trim_copy(request.set_shipping_address)}
    {}
    boost::optional<std::string> set_name;
    boost::optional<std::string> set_organization;
    boost::optional<PlaceAddress> set_place;
    boost::optional<PhoneNumber> set_telephone;
    boost::optional<PhoneNumber> set_fax;
    boost::optional<SetEmailAddresses> set_emails;
    boost::optional<SetEmailAddresses> set_notify_emails;
    boost::optional<VatIdentificationNumber> set_vat_identification_number;
    boost::optional<Util::Deletable<ContactAdditionalIdentifier>> set_additional_identifier;
    boost::optional<Util::Deletable<ContactAddress>> set_mailing_address;
    boost::optional<Util::Deletable<ContactAddress>> set_billing_address;
    std::array<boost::optional<Util::Deletable<ContactAddress>>, UpdateContactRequest::number_of_shipping_addresses> set_shipping_address;
};

struct IsIdentValueT : boost::static_visitor<bool>
{
    template <typename T>
    bool operator()(const T& value) const noexcept
    {
        return Fred::Registry::is_ident_value_t(get_raw_value_from(value));
    }
};
bool is_ident_value_t(const ContactAdditionalIdentifier& value) noexcept
{
    return boost::apply_visitor(IsIdentValueT{}, value);
}

template <typename T>
bool is_to_set(const boost::optional<T>& value)
{
    return value != boost::none;
}

template <typename T>
bool is_to_set(const boost::optional<Util::Deletable<T>>& deletable)
{
    return deletable != boost::none && is_to_set(get_raw_value_from(*deletable));
}

template <typename T, typename Check, typename Field>
void check_value(
        const boost::optional<T>& value,
        Field&& field,
        UpdateContactReply::Exception::InvalidData& invalid_data,
        Check&& is_ok)
{
    if (is_to_set(value))
    {
        if (!std::forward<Check>(is_ok)(*value))
        {
            invalid_data.fields.insert(std::forward<Field>(field));
        }
    }
}

template <typename T, typename Check, typename Field>
void check_value(
        const boost::optional<Util::Deletable<T>>& deletable,
        Field&& field,
        UpdateContactReply::Exception::InvalidData& invalid_data,
        Check&& is_ok)
{
    if (is_to_set(deletable))
    {
        check_value(get_raw_value_from(*deletable), std::forward<Field>(field), invalid_data, std::forward<Check>(is_ok));
    }
}

void check_update_contact_request(const LibFred::OperationContext& ctx, const UpdateContactRequest& request, const TrimmedChange& change)
{
    auto invalid_data = UpdateContactReply::Exception::InvalidData{request, {}};
    if (get_raw_value_from(request.by_registrar).empty())
    {
        invalid_data.fields.insert("by_registrar");
    }
    check_value(change.set_name, "set_name", invalid_data, [](auto&& value)
    {
        return Fred::Registry::is_postal_line_type(value);
    });
    check_value(change.set_organization, "set_organization", invalid_data, [](auto&& value)
    {
        return Fred::Registry::is_opt_postal_line_type(value);
    });
    const auto check_address = [&invalid_data, &ctx](const auto& address, auto make_address_name)
    {
        if (!Fred::Registry::is_country_code_valid(ctx, address.country_code))
        {
            invalid_data.fields.insert(make_address_name() + std::string{".country_code"});
        }
        if (address.street.empty() ||
            !Fred::Registry::is_postal_line_type(address.street[0]))
        {
            invalid_data.fields.insert(make_address_name() + std::string{".street[0]"});
        }
        static constexpr auto max_number_of_streets = 3;
        if (max_number_of_streets < address.street.size())
        {
            invalid_data.fields.insert(make_address_name() + std::string{".street.size()"});
        }
        for (std::size_t idx = 1; (idx < address.street.size()) && (idx < max_number_of_streets); ++idx)
        {
            if (!Fred::Registry::is_opt_postal_line_type(address.street[idx]))
            {
                invalid_data.fields.insert(make_address_name() + (".street[" + std::to_string(idx) + "]"));
            }
        }
        if (!Fred::Registry::is_postal_line_type(address.city))
        {
            invalid_data.fields.insert(make_address_name() + std::string{".city"});
        }
        if (!Fred::Registry::is_opt_postal_line_type(address.state_or_province))
        {
            invalid_data.fields.insert(make_address_name() + std::string{".state_or_province"});
        }
        if (!Fred::Registry::is_pc_type(address.postal_code))
        {
            invalid_data.fields.insert(make_address_name() + std::string{".postal_code"});
        }
    };
    if (is_to_set(change.set_place))
    {
        check_address(*change.set_place, []() { return "place"; });
    }
    if (is_to_set(change.set_mailing_address))
    {
        check_address(*get_raw_value_from(*change.set_mailing_address), []() { return "mailing_address"; });
    }
    if (is_to_set(change.set_billing_address))
    {
        check_address(*get_raw_value_from(*change.set_billing_address), []() { return "billing_address"; });
    }
    int address_idx = 0;
    std::for_each(begin(change.set_shipping_address), end(change.set_shipping_address), [&](auto&& deletable_address)
    {
        ++address_idx;
        if (is_to_set(deletable_address))
        {
            check_address(*get_raw_value_from(*deletable_address), [address_idx]() { return "shipping_address" + std::to_string(address_idx); });
        }
    });

    check_value(change.set_telephone, "set_telephone", invalid_data, [](auto&& value)
    {
        return Fred::Registry::is_e164_string_type(value);
    });
    check_value(change.set_fax, "set_fax", invalid_data, [](auto&& value)
    {
        return Fred::Registry::is_e164_string_type(value);
    });
    check_value(change.set_emails, "set_emails", invalid_data, [](auto&& emails)
    {
        return Fred::Registry::is_email_comma_list_type(emails.values);
    });
    check_value(change.set_notify_emails, "set_notify_emails", invalid_data, [](auto&& emails)
    {
        return Fred::Registry::is_email_comma_list_type(emails.values);
    });
    check_value(change.set_vat_identification_number, "set_vat_identification_number", invalid_data, [](auto&& value)
    {
        return Fred::Registry::is_vat_t(value);
    });
    check_value(change.set_additional_identifier, "set_additional_identifier", invalid_data, [](auto&& value)
    {
        return is_ident_value_t(value);
    });
    std::transform(
            begin(request.unused_discloseflags),
            end(request.unused_discloseflags),
            std::inserter(invalid_data.fields, end(invalid_data.fields)),
            [](auto&& discloseflag) { return "disclose[" + discloseflag + "]"; });
    if (request.log_entry_id != boost::none)
    {
        if (!fred_wrap(*request.log_entry_id).is_set())
        {
            invalid_data.fields.insert("log_entry_id");
        }
    }
    if (!invalid_data.fields.empty())
    {
        throw invalid_data;
    }
}

enum class ManualVerification
{
    unchanged,
    lost
};

auto conditionally_cancel_contact_states(const LibFred::OperationContext& ctx, unsigned long long contact_id)
{
    const auto type_of_object = Conversion::Enums::to_db_handle(LibFred::Object_Type::contact);
    const auto dbres = ctx.get_conn().exec_params(
            "SELECT (COALESCE(TRIM(c0.email), '') != COALESCE(TRIM(c1.email), '')) OR "
                   "(COALESCE(TRIM(c0.telephone), '') != COALESCE(TRIM(c1.telephone), '')) OR "
                   "(COALESCE(UPPER(TRIM(c0.name)), '') != COALESCE(UPPER(TRIM(c1.name)), '')) OR "
                   "(COALESCE(UPPER(TRIM(c0.organization)), '') != COALESCE(UPPER(TRIM(c1.organization)), '')) OR "
                   "(COALESCE(UPPER(TRIM(c0.street1)), '') != COALESCE(UPPER(TRIM(c1.street1)), '')) OR "
                   "(COALESCE(UPPER(TRIM(c0.street2)), '') != COALESCE(UPPER(TRIM(c1.street2)), '')) OR "
                   "(COALESCE(UPPER(TRIM(c0.street3)), '') != COALESCE(UPPER(TRIM(c1.street3)), '')) OR "
                   "(COALESCE(UPPER(TRIM(c0.city)), '') != COALESCE(UPPER(TRIM(c1.city)), '')) OR "
                   "(COALESCE(UPPER(TRIM(c0.stateorprovince)), '') != COALESCE(UPPER(TRIM(c1.stateorprovince)), '')) OR "
                   "(COALESCE(TRIM(c0.postalcode), '') != COALESCE(TRIM(c1.postalcode), '')) OR "
                   "(c0.country IS DISTINCT FROM c1.country) AS verification_loss, "
                   "(COALESCE(TRIM(c0.fax), '') != COALESCE(TRIM(c1.fax), '')) OR "
                   "(COALESCE(TRIM(c0.notifyemail), '') != COALESCE(TRIM(c1.notifyemail), '')) OR "
                   "(COALESCE(TRIM(c0.vat), '') != COALESCE(TRIM(c1.vat), '')) OR "
                   "(COALESCE(TRIM(c0.ssn), '') != COALESCE(TRIM(c1.ssn), '')) OR "
                   "(c0.ssntype IS DISTINCT FROM c1.ssntype) AS manual_verification_loss "
              "FROM object_registry obr "
              "JOIN contact c1 ON c1.id = obr.id "
              "JOIN history h0 ON h0.next = obr.historyid "
              "JOIN contact_history c0 ON c0.historyid = h0.id "
             "WHERE obr.id = $1::BIGINT AND "
                   "obr.type = get_object_type_id($2::TEXT) AND "
                   "obr.erdate IS NULL",
            Database::QueryParams{contact_id, type_of_object});
    if (dbres.size() != 1)
    {
        throw std::runtime_error{
                dbres.size() == 0 ? "unable to get contact data difference"
                                  : "ambiguous contact data"};
    }
    const auto verification_lost = static_cast<bool>(dbres[0][0]);
    const auto manual_verification_lost = verification_lost || static_cast<bool>(dbres[0][1]);
    const auto to_cancel = [&]()
    {
        if (verification_lost)
        {
            return LibFred::StatusList{
                LibFred::RegistrableObject::Contact::ConditionallyIdentifiedContact::name,
                LibFred::RegistrableObject::Contact::IdentifiedContact::name,
                LibFred::RegistrableObject::Contact::ContactPassedManualVerification::name,
                LibFred::RegistrableObject::Contact::ContactFailedManualVerification::name
            };
        }
        if (manual_verification_lost)
        {
            return LibFred::StatusList{
                LibFred::RegistrableObject::Contact::ContactPassedManualVerification::name,
                LibFred::RegistrableObject::Contact::ContactFailedManualVerification::name
            };
        }
        return LibFred::StatusList{};
    }();
    if (!to_cancel.empty())
    {
        try
        {
            LibFred::LockObjectStateRequestLock(contact_id).exec(ctx);
            LibFred::PerformObjectStateRequest(contact_id).exec(ctx);
            LibFred::CancelObjectStateRequestId{contact_id, to_cancel}.exec(ctx);
            LibFred::PerformObjectStateRequest(contact_id).exec(ctx);
            return manual_verification_lost ? ManualVerification::lost
                                            : ManualVerification::unchanged;
        }
        catch (const LibFred::CancelObjectStateRequestId::Exception& e)
        {
            if (!e.is_set_state_not_found() || e.is_set_object_id_not_found())
            {
                throw;
            }
            // swallow it - means that the state just wasn't set and nothing else
        }
    }
    return ManualVerification::unchanged;
}

class DbSavepoint
{
public:
    DbSavepoint(const LibFred::OperationContext& ctx, std::string name)
        : ctx_{ctx},
          name_{std::move(name)}
    {
        ctx_.get_conn().exec("SAVEPOINT " + name_);
    }
    DbSavepoint& release()
    {
        ctx_.get_conn().exec("RELEASE SAVEPOINT " + name_);
        return *this;
    }
    DbSavepoint& rollback()
    {
        ctx_.get_conn().exec("ROLLBACK TO " + name_);
        return *this;
    }
private:
    const LibFred::OperationContext& ctx_;
    std::string name_;
};

void update_contact_post_hooks(const LibFred::OperationContext& ctx, unsigned long long contact_id)
{
    DbSavepoint savepoint{ctx, "update_contact_post_hooks"};

    try
    {
        if (conditionally_cancel_contact_states(ctx, contact_id) == ManualVerification::lost)
        {
#if 0 // TODO: necessary when used as a substitute for EPP
            if (update_contact_enqueue_check)
            {
                enqueue_check_if_no_other_exists(
                        ctx,
                        contact_id,
                        LibFred::TestsuiteHandle::AUTOMATIC,
                        _logd_request_id);
            }
#endif
        }
        savepoint.release();
    }
    catch (...)
    {
        savepoint.rollback()
                 .release();
        throw;
    }
}

}//namespace Fred::Registry::Contact::{anonymous}

UpdateContactReply::Data update_contact(const UpdateContactRequest& request)
{
    try
    {
        FREDLOG_DEBUG("update_contact call");
        LibFred::OperationContextCreator ctx;
        LibFred::OperationContextLockingForShare ctx_for_share(ctx);
        LibFred::OperationContextLockingForUpdate ctx_for_update(ctx);
        const auto contact_uuid = fred_wrap(request.get_contact_id());
        const auto info_contact_output = LibFred::InfoContactByUuid{contact_uuid}.exec(ctx_for_update);
        const bool contact_is_registered = info_contact_output.info_contact_data.delete_time.isnull();
        if (!contact_is_registered)
        {
            throw UpdateContactReply::Exception::ContactDoesNotExist{request};
        }
        if (request.contact_snapshot_id != boost::none)
        {
            using CheckObjectChange = LibFred::RegistrableObject::CheckObjectChange;
            const auto check_result = CheckObjectChange{contact_uuid}
                .exec(ctx_for_update, fred_wrap(*request.contact_snapshot_id));
            if (check_result.data_has_been_changed || check_result.state_has_been_changed)
            {
                throw UpdateContactReply::Exception::InvalidSnapshot{request};
            }
        }
        // do it before any object state related checks
        LibFred::LockObjectStateRequestLock(info_contact_output.info_contact_data.id).exec(ctx);
        LibFred::PerformObjectStateRequest(info_contact_output.info_contact_data.id).exec(ctx);

        const auto contact_states_before_update = LibFred::RegistrableObject::Contact::GetContactStateById{info_contact_output.info_contact_data.id}.exec(ctx);
        const auto originator = get_raw_value_from(request.by_registrar);
        if (originator.empty())
        {
            throw UpdateContactReply::Exception::InvalidData{request, {"by_registrar"}};
        }
        const auto originator_data = LibFred::InfoRegistrarByHandle{originator}.exec(ctx_for_share).info_registrar_data;
        const bool is_system_registrar = originator_data.system.get_value_or(false);
        if (info_contact_output.info_contact_data.sponsoring_registrar_handle != originator)
        {
            const bool is_allowed_to_update = is_system_registrar;
            if (!is_allowed_to_update)
            {
                throw UpdateContactReply::Exception::UnauthorizedRegistrar{request};
            }
        }
        if (!is_system_registrar)
        {
            if (contact_states_before_update.are_set_any_of<LibFred::RegistrableObject::Contact::DeleteCandidate>())
            {
                throw UpdateContactReply::Exception::UpdateProhibited{request};
            }
            if (contact_states_before_update.are_set_any_of<LibFred::RegistrableObject::Contact::ServerUpdateProhibited>())
            {
                const auto is_mojeid_contact = contact_states_before_update.are_set_any_of<LibFred::RegistrableObject::Contact::MojeidContact>();
                if (!is_mojeid_contact)
                {
                    throw UpdateContactReply::Exception::UpdateProhibited{request};
                }
                FREDLOG_INFO("mojeID contact can be modified despite the serverUpdateProhibited flag");
            }
        }
        const auto trimmed_change = TrimmedChange{request};
        check_update_contact_request(ctx, request, trimmed_change);
        if (contact_states_before_update.are_set_all_of<LibFred::RegistrableObject::Contact::ServerContactNameChangeProhibited>() &&
            trimmed_change.set_name != boost::none)
        {
            throw UpdateContactReply::Exception::UpdateProhibited{request};
        }
        if (contact_states_before_update.are_set_all_of<LibFred::RegistrableObject::Contact::ServerContactOrganizationChangeProhibited>() &&
            trimmed_change.set_organization != boost::none)
        {
            throw UpdateContactReply::Exception::UpdateProhibited{request};
        }
        if (contact_states_before_update.are_set_all_of<LibFred::RegistrableObject::Contact::ServerContactIdentChangeProhibited>() &&
            trimmed_change.set_additional_identifier != boost::none)
        {
            throw UpdateContactReply::Exception::UpdateProhibited{request};
        }
        if (contact_states_before_update.are_set_all_of<LibFred::RegistrableObject::Contact::ServerContactPermanentAddressChangeProhibited>() &&
            trimmed_change.set_place != boost::none)
        {
            throw UpdateContactReply::Exception::UpdateProhibited{request};
        }
        LibFred::UpdateContactById update_contact_op{info_contact_output.info_contact_data.id, originator};
        optional_set(trimmed_change.set_name,
                     [&](const std::string& value) { update_contact_op.set_name(value); });
        optional_set(request.set_name_privacy,
                     [&](const Privacy& privacy) { update_contact_op.set_disclosename(privacy.is_publicly_available); });
        optional_set(trimmed_change.set_organization,
                     [&](const std::string& value) { update_contact_op.set_organization(value); });
        optional_set(request.set_organization_privacy,
                     [&](const Privacy& privacy) { update_contact_op.set_discloseorganization(privacy.is_publicly_available); });
        optional_set(trimmed_change.set_place,
                     [&](const PlaceAddress& address)
                     {
                         check_country_code(ctx, address, [&request](auto&&) { throw UpdateContactReply::Exception::InvalidData{request, {"set_place.country_code"}}; });
                         LibFred::Contact::PlaceAddress place;
                         if (0 < address.street.size())
                         {
                             place.street1 = address.street[0];
                             if (1 < address.street.size())
                             {
                                 place.street2 = address.street[1];
                                 if (2 < address.street.size())
                                 {
                                     place.street3 = address.street[2];
                                 }
                             }
                         }
                         place.city = address.city;
                         if (!address.state_or_province.empty())
                         {
                             place.stateorprovince = address.state_or_province;
                         }
                         place.postalcode = get_raw_value_from(address.postal_code);
                         place.country = get_raw_value_from(address.country_code);
                         update_contact_op.set_place(place);
                     });
        optional_set(request.set_place_privacy,
                     [&](const Privacy& privacy) { update_contact_op.set_discloseaddress(privacy.is_publicly_available); });
        optional_set(trimmed_change.set_telephone,
                     [&](const PhoneNumber& value) { update_contact_op.set_telephone(get_raw_value_from(value)); });
        optional_set(request.set_telephone_privacy,
                     [&](const Privacy& privacy) { update_contact_op.set_disclosetelephone(privacy.is_publicly_available); });
        optional_set(trimmed_change.set_fax,
                     [&](const PhoneNumber& value) { update_contact_op.set_fax(get_raw_value_from(value)); });
        optional_set(request.set_fax_privacy,
                     [&](const Privacy& privacy) { update_contact_op.set_disclosefax(privacy.is_publicly_available); });
        optional_set(trimmed_change.set_emails,
                     [&](const SetEmailAddresses& to_set)
                     {
                         if (to_set.values.empty())
                         {
                             throw NoEmail{};
                         }
                         update_contact_op.set_email(join_emails(to_set.values));
                     });
        optional_set(request.set_emails_privacy,
                     [&](const Privacy& privacy) { update_contact_op.set_discloseemail(privacy.is_publicly_available); });
        optional_set(trimmed_change.set_notify_emails,
                     [&](const SetEmailAddresses& to_set)
                     {
                         if (to_set.values.empty())
                         {
                             update_contact_op.set_notifyemail(Nullable<std::string>{});
                         }
                         else
                         {
                             update_contact_op.set_notifyemail(join_emails(to_set.values));
                         }
                     });
        optional_set(request.set_notify_emails_privacy,
                     [&](const Privacy& privacy) { update_contact_op.set_disclosenotifyemail(privacy.is_publicly_available); });
        optional_set(trimmed_change.set_vat_identification_number,
                     [&](const VatIdentificationNumber& value) { update_contact_op.set_vat(get_raw_value_from(value)); });
        optional_set(request.set_vat_identification_number_privacy,
                     [&](const Privacy& privacy) { update_contact_op.set_disclosevat(privacy.is_publicly_available); });
        const auto delete_personal_id = [&]()
        {
            update_contact_op.set_personal_id(Nullable<LibFred::PersonalIdUnion>{});
        };
        const auto set_personal_id = [&](const ContactAdditionalIdentifier& value)
        {
            class Visitor : public boost::static_visitor<void>
            {
            public:
                explicit Visitor(LibFred::UpdateContactById& update_contact_op) : update_contact_op_(update_contact_op) { }
                void operator()(const Contact::NationalIdentityNumber& src) const
                {
                    update_contact_op_.set_personal_id(LibFred::PersonalIdUnion::get_RC(get_raw_value_from(src)));
                }
                void operator()(const Contact::NationalIdentityCard& src) const
                {
                    update_contact_op_.set_personal_id(LibFred::PersonalIdUnion::get_OP(get_raw_value_from(src)));
                }
                void operator()(const Contact::PassportNumber& src) const
                {
                    update_contact_op_.set_personal_id(LibFred::PersonalIdUnion::get_PASS(get_raw_value_from(src)));
                }
                void operator()(const CompanyRegistrationNumber& src) const
                {
                    update_contact_op_.set_personal_id(LibFred::PersonalIdUnion::get_ICO(get_raw_value_from(src)));
                }
                void operator()(const Contact::SocialSecurityNumber& src) const
                {
                    update_contact_op_.set_personal_id(LibFred::PersonalIdUnion::get_MPSV(get_raw_value_from(src)));
                }
                void operator()(const Contact::Birthdate& src) const
                {
                    update_contact_op_.set_personal_id(LibFred::PersonalIdUnion::get_BIRTHDAY(get_raw_value_from(src)));
                }
            private:
                LibFred::UpdateContactById& update_contact_op_;
            };
            boost::apply_visitor(Visitor(update_contact_op), value);
        };
        select_operation(trimmed_change.set_additional_identifier, delete_personal_id, set_personal_id);
        optional_set(request.set_additional_identifier_privacy,
                     [&](const Privacy& privacy) { update_contact_op.set_discloseident(privacy.is_publicly_available); });
        select_operation(trimmed_change.set_mailing_address,
                         [&]() { update_contact_op.reset_address<LibFred::ContactAddressType::MAILING>(); },
                         [&](const ContactAddress& address_to_set)
                         {
                             check_country_code(ctx, address_to_set, [&request](auto&&)
                             {
                                 throw UpdateContactReply::Exception::InvalidData{request, {"set_mailing_address.country_code"}};
                             });
                             const auto libfred_address_to_set = to_libfred_contact_address(address_to_set);
                             update_contact_op.set_address<LibFred::ContactAddressType::MAILING>(libfred_address_to_set);
                         });
        select_operation(trimmed_change.set_billing_address,
                         [&]() { update_contact_op.reset_address<LibFred::ContactAddressType::BILLING>(); },
                         [&](const ContactAddress& address_to_set)
                         {
                             check_country_code(ctx, address_to_set, [&request](auto&&)
                             {
                                 throw UpdateContactReply::Exception::InvalidData{request, {"set_billing_address.country_code"}};
                             });
                             const auto libfred_address_to_set = to_libfred_contact_address(address_to_set);
                             update_contact_op.set_address<LibFred::ContactAddressType::BILLING>(libfred_address_to_set);
                         });
        select_operation(trimmed_change.set_shipping_address[0],
                         [&]() { update_contact_op.reset_address<LibFred::ContactAddressType::SHIPPING>(); },
                         [&](const ContactAddress& address_to_set)
                         {
                             check_country_code(ctx, address_to_set, [&request](auto&&)
                             {
                                 throw UpdateContactReply::Exception::InvalidData{request, {"set_shipping_address1.country_code"}};
                             });
                             const auto libfred_address_to_set = to_libfred_contact_address(address_to_set);
                             update_contact_op.set_address<LibFred::ContactAddressType::SHIPPING>(libfred_address_to_set);
                         });
        select_operation(trimmed_change.set_shipping_address[1],
                         [&]() { update_contact_op.reset_address<LibFred::ContactAddressType::SHIPPING_2>(); },
                         [&](const ContactAddress& address_to_set)
                         {
                             check_country_code(ctx, address_to_set, [&request](auto&&)
                             {
                                 throw UpdateContactReply::Exception::InvalidData{request, {"set_shipping_address2.country_code"}};
                             });
                             const auto libfred_address_to_set = to_libfred_contact_address(address_to_set);
                             update_contact_op.set_address<LibFred::ContactAddressType::SHIPPING_2>(libfred_address_to_set);
                         });
        select_operation(trimmed_change.set_shipping_address[2],
                         [&]() { update_contact_op.reset_address<LibFred::ContactAddressType::SHIPPING_3>(); },
                         [&](const ContactAddress& address_to_set)
                         {
                             check_country_code(ctx, address_to_set, [&request](auto&&)
                             {
                                 throw UpdateContactReply::Exception::InvalidData{request, {"set_shipping_address3.country_code"}};
                             });
                             const auto libfred_address_to_set = to_libfred_contact_address(address_to_set);
                             update_contact_op.set_address<LibFred::ContactAddressType::SHIPPING_3>(libfred_address_to_set);
                         });
        optional_set(request.set_warning_letter,
                     [&](const WarningLetter& warning_letter)
                     {
                         boost::optional<Nullable<bool>> to_set;
                         switch (warning_letter.preference)
                         {
                             case WarningLetter::SendingPreference::not_to_send:
                                 to_set = Nullable<bool>{false};
                                 break;
                             case WarningLetter::SendingPreference::to_send:
                                 to_set = Nullable<bool>{true};
                                 break;
                             case WarningLetter::SendingPreference::not_specified:
                                 to_set = Nullable<bool>{};
                                 break;
                         }
                         if (to_set == boost::none)
                         {
                             FREDLOG_WARNING("unknown value of set_warning_letter flag");
                         }
                         else
                         {
                             update_contact_op.set_domain_expiration_warning_letter_enabled(*to_set);
                         }
                     });
        const auto log_request_id = [&]()->boost::optional<unsigned long long>
        {
            if (request.log_entry_id == boost::none)
            {
                return boost::none;
            }
            const auto log_entry_id = fred_wrap(*request.log_entry_id);
            if (log_entry_id.is_set())
            {
                return log_entry_id.get_value();
            }
            FREDLOG_INFO("log_entry_id conversion failed");
            throw UpdateContactReply::Exception::InvalidData{request, {"log_entry_id"}};
        }();
        if (log_request_id != boost::none)
        {
            update_contact_op.set_logd_request_id(*log_request_id);
        }
        if (request.set_name_privacy != boost::none &&
             !request.set_name_privacy->is_publicly_available) // to hide name
        {
            throw UpdateContactReply::Exception::InvalidData{request, {"set_publish[\"name\"]"}};
        }
        if (request.set_organization_privacy != boost::none &&
             !request.set_organization_privacy->is_publicly_available) // to hide organization
        {
            throw UpdateContactReply::Exception::InvalidData{request, {"set_publish[\"organization\"]"}};
        }
        const bool to_remove_validation =
                contact_states_before_update.are_set_any_of<LibFred::RegistrableObject::Contact::IdentifiedContact,
                                                            LibFred::RegistrableObject::Contact::ValidatedContact>() &&
                (trimmed_change.set_name != boost::none ||
                 trimmed_change.set_organization != boost::none ||
                 trimmed_change.set_place != boost::none ||
                 trimmed_change.set_telephone != boost::none ||
                 trimmed_change.set_emails != boost::none);
        if (request.set_place_privacy != boost::none &&
            !request.set_place_privacy->is_publicly_available) // to hide address
        {
            if (contact_states_before_update.are_unset_all_of<LibFred::RegistrableObject::Contact::IdentifiedContact,
                                                              LibFred::RegistrableObject::Contact::ValidatedContact>() || // is not identified/validated
                to_remove_validation || // validation will be removed
                (trimmed_change.set_organization != boost::none &&
                 !trimmed_change.set_organization->empty()) || // organization has to be set
                (trimmed_change.set_organization == boost::none &&
                 !info_contact_output.info_contact_data.organization.get_value_or({}).empty())) // organization is set
            {
                // only identified/validated non-company contact can hide address
                throw UpdateContactReply::Exception::InvalidData{request, {"set_publish[\"place\"]"}};
            }
        }
        else if (!(request.set_place_privacy != boost::none && request.set_place_privacy->is_publicly_available) && // not to reveal address 
                 !info_contact_output.info_contact_data.discloseaddress && // address is hidden
                 trimmed_change.set_organization != boost::none &&
                 !trimmed_change.set_organization->empty()) // organization has to be set
        {
            // company cannot have hidden address
            throw UpdateContactReply::Exception::InvalidData{request, {"set_organization"}};
        }
        // perform UpdateContactById operation
        const auto numeric_history_id = update_contact_op.exec(ctx);
        update_contact_post_hooks(ctx, info_contact_output.info_contact_data.id);
        UpdateContactReply::Data result;
        result.contact_id = request.get_contact_id();
        result.contact_handle = Util::make_strong<ContactHandle>(info_contact_output.info_contact_data.handle);
        const auto after_update_data = LibFred::InfoContactHistoryByHistoryid(numeric_history_id).exec(ctx);
        result.contact_history_id = Util::make_strong<ContactHistoryId>(get_raw_value_from(after_update_data.info_contact_data.history_uuid));
        result.events = fred_unwrap(after_update_data.info_contact_data);
        LibFred::Poll::CreateUpdateOperationPollMessage<LibFred::Object_Type::contact>().exec(ctx, numeric_history_id);
        static constexpr auto no_log_request_id = 0ull;
        Notification::enqueue_notification(
                    ctx_for_update,
                    Notification::updated,
                    originator_data.id,
                    numeric_history_id,
                    ::Util::make_svtrid(log_request_id == boost::none ? no_log_request_id : *log_request_id));

        ctx.commit_transaction();
        return result;
    }
    catch (const InvalidEmail&)
    {
        struct InvalidEmail : public UpdateContactReply::Exception::InvalidData
        {
            explicit InvalidEmail(const UpdateContactRequest& request)
                : UpdateContactReply::Exception::InvalidData{request, {"set_emails", "set_notify_emails"}}
            { }
            const char* what() const noexcept override { return "invalid email format"; }
        };
        throw InvalidEmail{request};
    }
    catch (const NoEmail&)
    {
        struct EmailRequired : public UpdateContactReply::Exception::InvalidData
        {
            explicit EmailRequired(const UpdateContactRequest& request)
                : UpdateContactReply::Exception::InvalidData{request, {"set_emails"}}
            { }
            const char* what() const noexcept override { return "email is required"; }
        };
        throw EmailRequired{request};
    }
    catch (const UpdateContactReply::Exception::ContactDoesNotExist&)
    {
        throw;
    }
    catch (const UpdateContactReply::Exception::InvalidData&)
    {
        throw;
    }
    catch (const UpdateContactReply::Exception::InvalidSnapshot&)
    {
        throw;
    }
    catch (const UpdateContactReply::Exception::UpdateProhibited&)
    {
        throw;
    }
    catch (const UpdateContactReply::Exception::UnauthorizedRegistrar&)
    {
        throw;
    }
    catch (const LibFred::InfoContactByUuid::Exception& e)
    {
        if (e.is_set_unknown_contact_uuid())
        {
            throw UpdateContactReply::Exception::ContactDoesNotExist(request);
        }
        throw Registry::InternalServerError("unexpected problem signalized by LibFred::InfoContactByUuid::Exception");
    }
    catch (const LibFred::InfoRegistrarByHandle::Exception& e)
    {
        if (e.is_set_unknown_registrar_handle())
        {
            throw UpdateContactReply::Exception::RegistrarDoesNotExist{request};
        }
        FREDLOG_INFO(std::string{"LibFred::InfoRegistrarByHandle::Exception: "} + e.what());
        throw Registry::InternalServerError("unexpected problem signalized by LibFred::InfoRegistrarByHandle::Exception");
    }
    catch (const LibFred::UpdateContactById::ExceptionType& e)
    {
        if (e.is_set_unknown_country())
        {
            struct InvalidCountry : public UpdateContactReply::Exception::InvalidData
            {
                explicit InvalidCountry(const UpdateContactRequest& request)
                    : UpdateContactReply::Exception::InvalidData{request, {"set_place.country_code"}}
                { }
                const char* what() const noexcept override { return "invalid country"; }
            };
            throw InvalidCountry{request};
        }
        if (e.is_set_forbidden_company_name_setting())
        {
            struct ForbiddenCompanyNameSetting : public UpdateContactReply::Exception::InvalidData
            {
                explicit ForbiddenCompanyNameSetting(const UpdateContactRequest& request)
                    : UpdateContactReply::Exception::InvalidData{request, {"set_organization"}}
                { }
                const char* what() const noexcept override { return "forbidden company name setting"; }
            };
            throw ForbiddenCompanyNameSetting{request};
        }
        if (e.is_set_unknown_registrar_handle())
        {
            throw UpdateContactReply::Exception::RegistrarDoesNotExist{request};
        }
        FREDLOG_INFO(std::string{"LibFred::UpdateContactById::ExceptionType: "} + e.what());
        throw Registry::InternalServerError("unexpected problem signalized by LibFred::UpdateContactById::ExceptionType");
    }
    catch (const LibFred::UnknownObjectId&)
    {
        throw UpdateContactReply::Exception::ContactDoesNotExist(request);
    }
    catch (const Database::Exception& e)
    {
        FREDLOG_INFO(std::string{"Database::Exception: "} + e.what());
        throw Registry::DatabaseError(std::string("Database::Exception: ") + e.what());
    }
    catch (const std::exception& e)
    {
        FREDLOG_INFO(std::string{"std::exception: "} + e.what());
        throw Registry::InternalServerError("unexpected unknown exception");
    }
    catch (...)
    {
        FREDLOG_INFO("unknown exception");
        throw Registry::InternalServerError("unexpected unknown exception");
    }
}

}//namespace Fred::Registry::Contact
}//namespace Fred::Registry
}//namespace Fred
