/*
 * Copyright (C) 2019-2025  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/domain/domain_state.hh"

#include "src/fred_unwrap.hh"
#include "src/fred_wrap.hh"
#include "src/util/strong_type.hh"
#include "src/util/struct_to_string.hh"

#include "libfred/db_settings.hh"
#include "libfred/registrable_object/domain/get_domain_state.hh"
#include "libfred/registrable_object/snapshot_id.hh"
#include "util/db/db_exceptions.hh"

#include <boost/uuid/string_generator.hpp>

#include <utility>

namespace Fred {
namespace Registry {
namespace Domain {

DomainId DomainStateRequest::get_domain_id() const
{
    try
    {
        return Util::make_strong<DomainId>(boost::uuids::string_generator{}(domain_id));
    }
    catch (...)
    {
        throw DomainStateReply::Exception::InvalidData{*this, {"domain_id"}};
    }
}

std::string DomainStateRequest::to_string() const
{
    return Util::StructToString().add("domain_id", domain_id)
                                 .add("domain_snapshot_id", domain_snapshot_id)
                                 .finish();
}

std::string DomainState::to_string() const
{
    if (flag_presents.empty())
    {
        return "{ }";
    }
    std::string buffer;
    for (const auto& presents : flag_presents)
    {
        if (!buffer.empty())
        {
            buffer += ", ";
        }
        buffer += "'" + presents.first + "': " + (presents.second ? "presents" : "absents");
    }
    return "{ " + buffer + " }";
}

std::string DomainStateReply::Data::to_string() const
{
    return Util::StructToString().add("domain_id", domain_id)
                                 .add("state", state)
                                 .add("snapshot_id", snapshot_id)
                                 .finish();
}

DomainStateReply::Exception::DomainDoesNotExist::DomainDoesNotExist(const DomainStateRequest& src)
    : Registry::Exception("domain " + src.to_string() + " does not exist")
{
}

DomainStateReply::Exception::InvalidSnapshot::InvalidSnapshot(const DomainStateRequest& src)
    : Registry::Exception{"snapshot_id of domain " + Util::Into<std::string>::from(src.domain_id) + " "
                          "differs from current version"}
{ }

DomainStateReply::Exception::InvalidData::InvalidData(const DomainStateRequest& src, std::set<std::string> fields)
    : Registry::Exception{"invalid data: " + src.to_string()},
      fields{std::move(fields)}
{ }

DomainStateReply::Data domain_state(const DomainStateRequest& request)
{
    try
    {
        LibFred::OperationContextCreator ctx;
        LibFred::OperationContextLockingForShare ctx_for_share{ctx};
        DomainStateReply::Data result;
        const auto domain_id = request.get_domain_id();
        const auto domain_uuid = fred_wrap(domain_id);
        if (request.domain_snapshot_id != boost::none)
        {
            try
            {
                using CheckObjectChange = LibFred::RegistrableObject::CheckObjectChange;
                const auto check_result = CheckObjectChange{domain_uuid}
                    .exec(ctx_for_share, fred_wrap(*request.domain_snapshot_id));
                if (check_result.data_has_been_changed || check_result.state_has_been_changed)
                {
                    throw DomainStateReply::Exception::InvalidSnapshot{request};
                }
            }
            catch (const LibFred::RegistrableObject::InvalidSnapshotId&)
            {
                throw DomainStateReply::Exception::InvalidData{request, {"domain_snapshot_id"}};
            }
            catch (const LibFred::UnknownObjectId&)
            {
                throw DomainStateReply::Exception::DomainDoesNotExist(request);
            }
        }
        result.state = fred_unwrap(LibFred::RegistrableObject::Domain::GetDomainStateByUuid(domain_uuid).exec(ctx_for_share));
        result.domain_id = domain_id;
        result.snapshot_id = fred_unwrap(LibFred::RegistrableObject::get_snapshot_id(ctx));
        ctx.commit_transaction();
        return result;
    }
    catch (const LibFred::RegistrableObject::Domain::GetDomainStateByUuid::DoesNotExist&)
    {
        throw DomainStateReply::Exception::DomainDoesNotExist(request);
    }
    catch (const DomainStateReply::Exception::DomainDoesNotExist&)
    {
        throw;
    }
    catch (const DomainStateReply::Exception::InvalidData&)
    {
        throw;
    }
    catch (const DomainStateReply::Exception::InvalidSnapshot&)
    {
        throw;
    }
    catch (const Database::Exception& e)
    {
        throw Registry::DatabaseError(std::string("Database::Exception: ") + e.what());
    }
    catch (const std::exception& e)
    {
        throw Registry::InternalServerError(std::string("expected unknown exception: ") + e.what());
    }
    catch (...)
    {
        throw Registry::InternalServerError("unexpected unknown exception");
    }
}

} // namespace Fred::Registry::Domain
} // namespace Fred::Registry
} // namespace Fred
