/*
 * Copyright (C) 2025  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/keyset/keyset_authinfo.hh"

#include "src/cfg.hh"
#include "src/fred_unwrap.hh"
#include "src/util/into.hh"
#include "src/util/strong_type.hh"
#include "src/util/struct_to_string.hh"

#include "libfred/object/clean_authinfo.hh"
#include "libfred/object/generate_authinfo_password.hh"
#include "libfred/object/store_authinfo.hh"
#include "libfred/registrable_object/keyset/info_keyset.hh"
#include "libfred/registrable_object/is_registered.hh"
#include "libfred/registrable_object/snapshot_id.hh"
#include "libfred/registrar/info_registrar.hh"
#include "util/db/db_exceptions.hh"

#include <utility>

namespace Fred {
namespace Registry {
namespace Keyset {

std::string AddKeysetAuthInfoRequest::to_string() const
{
    return Util::StructToString().add("keyset_id", keyset_id)
                                 .add("keyset_snapshot_id", keyset_snapshot_id)
                                 .add("auth_info", std::string{"********"}) // do not reveal a secret
                                 .add("ttl", ttl)
                                 .add("by_registrar", by_registrar)
                                 .add("log_entry_id", log_entry_id)
                                 .finish();
}

KeysetId AddKeysetAuthInfoRequest::get_keyset_id() const
{
    try
    {
        return Util::make_strong<KeysetId>(boost::uuids::string_generator{}(keyset_id));
    }
    catch (...)
    {
        throw AddKeysetAuthInfoReply::Exception::InvalidData{*this, {"keyset_id"}};
    }
}

AddKeysetAuthInfoReply::Data::Data()
    : generated_auth_info{boost::none}
{ }

AddKeysetAuthInfoReply::Data::Data(std::string password)
    : generated_auth_info{std::move(password)}
{ }

std::string AddKeysetAuthInfoReply::Data::to_string() const
{
    return Util::StructToString().add("generated_auth_info", std::string{generated_auth_info == boost::none ? "none" : "********"}) // do not reveal a secret
                                 .finish();
}

AddKeysetAuthInfoReply::Exception::KeysetDoesNotExist::KeysetDoesNotExist(const AddKeysetAuthInfoRequest& src)
    : Registry::Exception{"keyset " + src.keyset_id + " does not exist"}
{ }

AddKeysetAuthInfoReply::Exception::RegistrarDoesNotExist::RegistrarDoesNotExist(const AddKeysetAuthInfoRequest& src)
    : Registry::Exception{"registrar " + get_raw_value_from(src.by_registrar) + " does not exist"}
{ }

AddKeysetAuthInfoReply::Exception::InvalidData::InvalidData(const AddKeysetAuthInfoRequest& src, std::set<std::string> fields)
    : Registry::Exception{"invalid data: " + src.to_string()},
      fields{std::move(fields)}
{ }

AddKeysetAuthInfoReply::Exception::InvalidSnapshot::InvalidSnapshot(const AddKeysetAuthInfoRequest& src)
    : Registry::Exception{"keyset " + src.keyset_id + " differs from snapshot " + Util::Into<std::string>::from(src.keyset_snapshot_id)}
{ }

namespace {

auto get_registrar_id(const LibFred::OperationContext& ctx, const AddKeysetAuthInfoRequest& request)
{
    const auto registrar_handle = get_raw_value_from(request.by_registrar);
    if (registrar_handle.empty())
    {
        throw AddKeysetAuthInfoReply::Exception::InvalidData{request, {"by_registrar"}};
    }
    try
    {
        return LibFred::InfoRegistrarByHandle{registrar_handle}.exec(ctx).info_registrar_data.id;
    }
    catch (const LibFred::InfoRegistrarByHandle::Exception& e)
    {
        if (e.is_set_unknown_registrar_handle())
        {
            throw AddKeysetAuthInfoReply::Exception::RegistrarDoesNotExist{request};
        }
        throw;
    }
}

template <LibFred::DbLock lock>
auto get_object_id(const LibFred::OperationContextUsing<lock>& ctx, const KeysetId& id)
{
    const auto keyset_uuid = LibFred::RegistrableObject::make_uuid_of<LibFred::Object_Type::keyset>(get_raw_value_from(id));
    return LibFred::Object::ObjectId{LibFred::InfoKeysetByUuid{keyset_uuid}.exec(ctx).info_keyset_data.id};
}

LibFred::RegistrableObject::Keyset::KeysetUuid make_libfred_keyset_uuid(const KeysetId& src)
{
    return LibFred::RegistrableObject::make_uuid_of<LibFred::Object_Type::keyset>(get_raw_value_from(src));
}

LibFred::RegistrableObject::SnapshotId make_libfred_snapshot_id(const SnapshotId& src)
{
    return ::Util::make_strong<LibFred::RegistrableObject::SnapshotId>(get_raw_value_from(src));
}

}//namespace Fred::Registry::Keyset::{anonymous}

}//namespace Fred::Registry::Keyset
}//namespace Fred::Registry
}//namespace Fred

using namespace Fred::Registry::Keyset;

AddKeysetAuthInfoReply::Data Fred::Registry::Keyset::add_keyset_auth_info(const AddKeysetAuthInfoRequest& request)
{
    try
    {
        LibFred::OperationContextCreator ctx;
        try
        {
            const auto result = [&request]()
            {
                if (request.auth_info != boost::none)
                {
                    return AddKeysetAuthInfoReply::Data{};
                }
                return AddKeysetAuthInfoReply::Data{LibFred::generate_authinfo_pw().password_};
            }();
            const auto& password = [&request, &result]()
            {
                if (request.auth_info != boost::none)
                {
                    return *request.auth_info;
                }
                return *result.generated_auth_info;
            }();
            const auto ttl = [&request]()
            {
                if (request.ttl != boost::none)
                {
                    return *request.ttl;
                }
                return Cfg::Options::get().fred.auth_info_ttl;
            }();
            LibFred::OperationContextLockingForShare ctx_for_share{ctx};
            if (request.keyset_snapshot_id != boost::none)
            {
                using CheckObjectChange = LibFred::RegistrableObject::CheckObjectChange;
                const auto check_result = CheckObjectChange{make_libfred_keyset_uuid(request.get_keyset_id())}
                    .exec(ctx_for_share, make_libfred_snapshot_id(*request.keyset_snapshot_id));
                if (check_result.data_has_been_changed || check_result.state_has_been_changed)
                {
                    throw AddKeysetAuthInfoReply::Exception::InvalidSnapshot{request};
                }
            }
            const auto store_authinfo = LibFred::Object::StoreAuthinfo{
                    get_object_id(ctx_for_share, request.get_keyset_id()),
                    get_registrar_id(ctx_for_share, request),
                    ttl};
            store_authinfo.exec(ctx, password);
            ctx.commit_transaction();
            return result;
        }
        catch (const LibFred::InfoKeysetByUuid::Exception& e)
        {
            if (e.is_set_unknown_keyset_uuid())
            {
                if (LibFred::is_registered_by_uuid(ctx, LibFred::Object_Type::keyset, e.get_unknown_keyset_uuid()))
                {
                    // Issue fred/libfred#30 workaround
                    throw Registry::DatabaseError{"Database::Exception: failed to get info about registered keyset"};
                }
                throw AddKeysetAuthInfoReply::Exception::KeysetDoesNotExist{request};
            }
            throw Registry::InternalServerError{"unexpected problem signalized by LibFred::InfoKeysetByUuid::Exception"};
        }
        catch (const LibFred::UnknownObjectId&)
        {
            throw AddKeysetAuthInfoReply::Exception::KeysetDoesNotExist{request};
        }
    }
    catch (const LibFred::Object::InvalidTtl&)
    {
        throw AddKeysetAuthInfoReply::Exception::InvalidData{request, {"ttl"}};
    }
    catch (const LibFred::RegistrableObject::InvalidSnapshotId&)
    {
        throw AddKeysetAuthInfoReply::Exception::InvalidData{request, {"keyset_snapshot_id"}};
    }
    catch (const AddKeysetAuthInfoReply::Exception::KeysetDoesNotExist&)
    {
        throw;
    }
    catch (const AddKeysetAuthInfoReply::Exception::RegistrarDoesNotExist&)
    {
        throw;
    }
    catch (const AddKeysetAuthInfoReply::Exception::InvalidData&)
    {
        throw;
    }
    catch (const AddKeysetAuthInfoReply::Exception::InvalidSnapshot&)
    {
        throw;
    }
    catch (const Registry::DatabaseError&)
    {
        throw;
    }
    catch (const Registry::InternalServerError&)
    {
        throw;
    }
    catch (const Database::Exception& e)
    {
        throw Registry::DatabaseError{std::string("Database::Exception: ") + e.what()};
    }
    catch (const std::exception& e)
    {
        throw Registry::InternalServerError{std::string{"expected unknown exception: "} + e.what()};
    }
    catch (...)
    {
        throw Registry::InternalServerError("unexpected unknown exception");
    }
}
