/*
 * Copyright (C) 2019-2025  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/keyset/keyset_state.hh"

#include "src/fred_unwrap.hh"
#include "src/fred_wrap.hh"
#include "src/util/strong_type.hh"
#include "src/util/struct_to_string.hh"

#include "libfred/db_settings.hh"
#include "libfred/registrable_object/keyset/get_keyset_state.hh"
#include "libfred/registrable_object/snapshot_id.hh"
#include "util/db/db_exceptions.hh"

#include <boost/uuid/string_generator.hpp>

#include <utility>

namespace Fred {
namespace Registry {
namespace Keyset {

KeysetId KeysetStateRequest::get_keyset_id() const
{
    try
    {
        return Util::make_strong<KeysetId>(boost::uuids::string_generator{}(keyset_id));
    }
    catch (...)
    {
        throw KeysetStateReply::Exception::InvalidData{*this, {"keyset_id"}};
    }
}

std::string KeysetStateRequest::to_string() const
{
    return Util::StructToString().add("keyset_id", keyset_id)
                                 .add("keyset_snapshot_id", keyset_snapshot_id)
                                 .finish();
}

std::string KeysetState::to_string() const
{
    if (flag_presents.empty())
    {
        return "{ }";
    }
    std::string buffer;
    for (const auto& presents : flag_presents)
    {
        if (!buffer.empty())
        {
            buffer += ", ";
        }
        buffer += "'" + presents.first + "': " + (presents.second ? "presents" : "absents");
    }
    return "{ " + buffer + " }";
}

std::string KeysetStateReply::Data::to_string() const
{
    return Util::StructToString().add("keyset_id", keyset_id)
                                 .add("state", state)
                                 .add("snapshot_id", snapshot_id)
                                 .finish();
}

KeysetStateReply::Exception::KeysetDoesNotExist::KeysetDoesNotExist(const KeysetStateRequest& src)
    : Registry::Exception("keyset " + src.to_string() + " does not exist")
{
}

KeysetStateReply::Exception::InvalidSnapshot::InvalidSnapshot(const KeysetStateRequest& src)
    : Registry::Exception{"snapshot_id of keyset " + Util::Into<std::string>::from(src.keyset_id) + " "
                          "differs from current version"}
{ }

KeysetStateReply::Exception::InvalidData::InvalidData(const KeysetStateRequest& src, std::set<std::string> fields)
    : Registry::Exception{"invalid data: " + src.to_string()},
      fields{std::move(fields)}
{ }

KeysetStateReply::Data keyset_state(const KeysetStateRequest& request)
{
    try
    {
        LibFred::OperationContextCreator ctx;
        LibFred::OperationContextLockingForShare ctx_for_share{ctx};
        KeysetStateReply::Data result;
        const auto keyset_id = request.get_keyset_id();
        const auto keyset_uuid = fred_wrap(keyset_id);
        if (request.keyset_snapshot_id != boost::none)
        {
            try
            {
                using CheckObjectChange = LibFred::RegistrableObject::CheckObjectChange;
                const auto check_result = CheckObjectChange{keyset_uuid}
                    .exec(ctx_for_share, fred_wrap(*request.keyset_snapshot_id));
                if (check_result.data_has_been_changed || check_result.state_has_been_changed)
                {
                    throw KeysetStateReply::Exception::InvalidSnapshot{request};
                }
            }
            catch (const LibFred::RegistrableObject::InvalidSnapshotId&)
            {
                throw KeysetStateReply::Exception::InvalidData{request, {"keyset_snapshot_id"}};
            }
            catch (const LibFred::UnknownObjectId&)
            {
                throw KeysetStateReply::Exception::KeysetDoesNotExist{request};
            }
        }
        result.state = fred_unwrap(LibFred::RegistrableObject::Keyset::GetKeysetStateByUuid(keyset_uuid).exec(ctx_for_share));
        result.keyset_id = keyset_id;
        result.snapshot_id = fred_unwrap(LibFred::RegistrableObject::get_snapshot_id(ctx));
        ctx.commit_transaction();
        return result;
    }
    catch (const LibFred::RegistrableObject::Keyset::GetKeysetStateByHandle::DoesNotExist&)
    {
        throw KeysetStateReply::Exception::KeysetDoesNotExist(request);
    }
    catch (const KeysetStateReply::Exception::InvalidData&)
    {
        throw;
    }
    catch (const KeysetStateReply::Exception::InvalidSnapshot&)
    {
        throw;
    }
    catch (const KeysetStateReply::Exception::KeysetDoesNotExist&)
    {
        throw;
    }
    catch (const Database::Exception& e)
    {
        throw Registry::DatabaseError(std::string("Database::Exception: ") + e.what());
    }
    catch (const std::exception& e)
    {
        throw Registry::InternalServerError(std::string("expected unknown exception: ") + e.what());
    }
    catch (...)
    {
        throw Registry::InternalServerError("unexpected unknown exception");
    }
}

} // namespace Fred::Registry::Keyset
} // namespace Fred::Registry
} // namespace Fred
