/*
 * Copyright (C) 2019-2025  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

 #include "src/nsset/nsset_state.hh"

#include "src/fred_unwrap.hh"
#include "src/fred_wrap.hh"
#include "src/util/strong_type.hh"
#include "src/util/struct_to_string.hh"

#include "libfred/db_settings.hh"
#include "libfred/registrable_object/nsset/get_nsset_state.hh"
#include "libfred/registrable_object/snapshot_id.hh"
#include "util/db/db_exceptions.hh"

#include <boost/uuid/string_generator.hpp>

#include <utility>

namespace Fred {
namespace Registry {
namespace Nsset {

NssetId NssetStateRequest::get_nsset_id() const
{
    try
    {
        return Util::make_strong<NssetId>(boost::uuids::string_generator{}(nsset_id));
    }
    catch (...)
    {
        throw NssetStateReply::Exception::InvalidData{*this, {"nsset_id"}};
    }
}

std::string NssetStateRequest::to_string() const
{
    return Util::StructToString().add("nsset_id", nsset_id)
                                 .add("nsset_snapshot_id", nsset_snapshot_id)
                                 .finish();
}

std::string NssetState::to_string() const
{
    if (flag_presents.empty())
    {
        return "{ }";
    }
    std::string buffer;
    for (const auto& presents : flag_presents)
    {
        if (!buffer.empty())
        {
            buffer += ", ";
        }
        buffer += "'" + presents.first + "': " + (presents.second ? "presents" : "absents");
    }
    return "{ " + buffer + " }";
}

std::string NssetStateReply::Data::to_string() const
{
    return Util::StructToString().add("nsset_id", nsset_id)
                                 .add("state", state)
                                 .add("snapshot_id", snapshot_id)
                                 .finish();
}

NssetStateReply::Exception::NssetDoesNotExist::NssetDoesNotExist(const NssetStateRequest& src)
    : Registry::Exception("nsset " + src.to_string() + " does not exist")
{
}

NssetStateReply::Exception::InvalidSnapshot::InvalidSnapshot(const NssetStateRequest& src)
    : Registry::Exception{"snapshot_id of nsset " + Util::Into<std::string>::from(src.nsset_id) + " "
                          "differs from current version"}
{ }

NssetStateReply::Exception::InvalidData::InvalidData(const NssetStateRequest& src, std::set<std::string> fields)
    : Registry::Exception{"invalid data: " + src.to_string()},
      fields{std::move(fields)}
{ }

NssetStateReply::Data nsset_state(const NssetStateRequest& request)
{
    try
    {
        LibFred::OperationContextCreator ctx;
        LibFred::OperationContextLockingForShare ctx_for_share{ctx};
        NssetStateReply::Data result;
        const auto nsset_id = request.get_nsset_id();
        const auto nsset_uuid = fred_wrap(nsset_id);
        if (request.nsset_snapshot_id != boost::none)
        {
            try
            {
                using CheckObjectChange = LibFred::RegistrableObject::CheckObjectChange;
                const auto check_result = CheckObjectChange{nsset_uuid}
                    .exec(ctx_for_share, fred_wrap(*request.nsset_snapshot_id));
                if (check_result.data_has_been_changed || check_result.state_has_been_changed)
                {
                    throw NssetStateReply::Exception::InvalidSnapshot{request};
                }
            }
            catch (const LibFred::RegistrableObject::InvalidSnapshotId&)
            {
                throw NssetStateReply::Exception::InvalidData{request, {"nsset_snapshot_id"}};
            }
            catch (const LibFred::UnknownObjectId&)
            {
                throw NssetStateReply::Exception::NssetDoesNotExist{request};
            }
        }
        result.state = fred_unwrap(LibFred::RegistrableObject::Nsset::GetNssetStateByUuid(nsset_uuid).exec(ctx_for_share));
        result.nsset_id = nsset_id;
        result.snapshot_id = fred_unwrap(LibFred::RegistrableObject::get_snapshot_id(ctx));
        ctx.commit_transaction();
        return result;
    }
    catch (const LibFred::RegistrableObject::Nsset::GetNssetStateByHandle::DoesNotExist&)
    {
        throw NssetStateReply::Exception::NssetDoesNotExist(request);
    }
    catch (const NssetStateReply::Exception::InvalidData&)
    {
        throw;
    }
    catch (const NssetStateReply::Exception::InvalidSnapshot&)
    {
        throw;
    }
    catch (const NssetStateReply::Exception::NssetDoesNotExist&)
    {
        throw;
    }
    catch (const Database::Exception& e)
    {
        throw Registry::DatabaseError(std::string("Database::Exception: ") + e.what());
    }
    catch (const std::exception& e)
    {
        throw InternalServerError(std::string("expected unknown exception: ") + e.what());
    }
    catch (...)
    {
        throw InternalServerError("unexpected unknown exception");
    }
}

} // namespace Fred::Registry::Nsset
} // namespace Fred::Registry
} // namespace Fred
