/*
 * Copyright (C) 2025  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/nsset/update_nsset_state.hh"
#include "src/nsset/nsset_state_flags.hh"

#include "src/exceptions.hh"
#include "src/fred_unwrap.hh"
#include "src/util/into.hh"
#include "src/util/struct_to_string.hh"

#include "libfred/object_state/cancel_object_state_request_id.hh"
#include "libfred/object_state/create_object_state_request_id.hh"
#include "libfred/object_state/perform_object_state_request.hh"
#include "libfred/opcontext.hh"
#include "libfred/registrable_object/nsset/info_nsset.hh"
#include "libfred/registrable_object/is_registered.hh"
#include "libfred/registrable_object/snapshot_id.hh"
#include "util/db/db_exceptions.hh"

#include <algorithm>
#include <set>

namespace Fred {
namespace Registry {
namespace Nsset {

std::string UpdateNssetStateRequest::to_string()const
{
    return Util::StructToString().add("updates", updates)
                                 .add("nsset_snapshot_id", nsset_snapshot_id)
                                 .add("log_entry_id", log_entry_id)
                                 .finish();
}

bool UpdateNssetStateRequest::NssetIdLess::operator()(const NssetId& lhs, const NssetId& rhs) const noexcept
{
    return get_raw_value_from(lhs) < get_raw_value_from(rhs);
}

UpdateNssetStateReply::Exception::NssetDoesNotExist::NssetDoesNotExist(const UpdateNssetStateRequest&)
    : Registry::Exception{"at least one nsset does not exist"}
{ }

UpdateNssetStateReply::Exception::InvalidSnapshot::InvalidSnapshot(const UpdateNssetStateRequest&)
    : Registry::Exception{"at least one nsset has been changed"}
{ }

UpdateNssetStateReply::Exception::StateFlagDoesNotExist::StateFlagDoesNotExist(const UpdateNssetStateRequest&)
    : Registry::Exception{"at least one state flag does not exist"}
{ }

UpdateNssetStateReply::Exception::InvalidData::InvalidData(const UpdateNssetStateRequest&, std::set<std::string> fields)
    : Registry::Exception{"invalid data"},
      fields{std::move(fields)}
{ }

}//namespace Fred::Registry::Nsset
}//namespace Fred::Registry
}//namespace Fred

using namespace Fred::Registry::Nsset;

namespace {

void existing_manual_state_flags_required(const UpdateNssetStateRequest& request, const std::map<std::string, bool>& set_flags)
{
    const auto manual_flags = []()
    {
        const auto flags_info = get_nsset_state_flags_info();
        std::set<std::string> result;
        std::for_each(begin(flags_info), end(flags_info), [&result](auto&& info)
        {
            if (info.how_to_set == StateFlagInfo::Manipulation::manual)
            {
                result.insert(info.name);
            }
        });
        return result;
    }();
    const auto flag_is_manual = [&manual_flags](auto&& name_action_pair)
    {
        return manual_flags.find(name_action_pair.first) != end(manual_flags);
    };
    if (!std::all_of(begin(set_flags), end(set_flags), flag_is_manual))
    {
        throw UpdateNssetStateReply::Exception::StateFlagDoesNotExist{request};
    }
}

LibFred::RegistrableObject::SnapshotId make_libfred_snapshot_id(const Fred::Registry::SnapshotId& src)
{
    return ::Util::make_strong<LibFred::RegistrableObject::SnapshotId>(get_raw_value_from(src));
}

void update_state(
        const LibFred::OperationContextLockingForUpdate& ctx,
        const UpdateNssetStateRequest& request,
        const NssetId& nsset_id,
        const UpdateNssetStateRequest::SetFlags& set_flags)
{
    const auto nsset_uuid = LibFred::RegistrableObject::make_uuid_of<LibFred::Object_Type::nsset>(get_raw_value_from(nsset_id));
    const auto info_nsset_data = [&ctx, &nsset_uuid, &request]()
    {
        try
        {
            return LibFred::InfoNssetByUuid{nsset_uuid}.exec(ctx).info_nsset_data;
        }
        catch (const LibFred::InfoNssetByUuid::Exception& e)
        {
            if (e.is_set_unknown_nsset_uuid())
            {
                if (LibFred::is_registered_by_uuid(ctx, LibFred::Object_Type::nsset, e.get_unknown_nsset_uuid()))
                {
                    // Issue fred/libfred#30 workaround
                    throw Fred::Registry::DatabaseError{"Database::Exception: failed to get info about registered nsset"};
                }
                throw UpdateNssetStateReply::Exception::NssetDoesNotExist{request};
            }
            throw Fred::Registry::InternalServerError{"unexpected problem signalized by LibFred::InfoNssetByUuid::Exception"};
        }
    }();
    const bool nsset_is_registered = info_nsset_data.delete_time.isnull();
    if (!nsset_is_registered)
    {
        throw UpdateNssetStateReply::Exception::NssetDoesNotExist{request};
    }
    if (request.nsset_snapshot_id != boost::none)
    {
        using CheckObjectChange = LibFred::RegistrableObject::CheckObjectChange;
        const auto check_result = CheckObjectChange{nsset_uuid}
            .exec(ctx, make_libfred_snapshot_id(*request.nsset_snapshot_id));
        if (check_result.data_has_been_changed ||
            check_result.state_has_been_changed)
        {
            throw UpdateNssetStateReply::Exception::InvalidSnapshot{request};
        }
    }
    existing_manual_state_flags_required(request, set_flags);
    bool request_changed = false;
    std::set<std::string> flags_to_set;
    std::for_each(begin(set_flags), end(set_flags), [&](auto&& name_action_pair)
    {
        const auto flag = name_action_pair.first;
        const auto to_set = name_action_pair.second;
        try
        {
            LibFred::CancelObjectStateRequestId{info_nsset_data.id, {flag}}.exec(ctx);
            request_changed = true;
        }
        catch (const LibFred::CancelObjectStateRequestId::Exception& e)
        {
            if (e.is_set_object_id_not_found())
            {
                throw UpdateNssetStateReply::Exception::NssetDoesNotExist{request};
            }
            const bool allowed_problem_occured = e.is_set_state_not_found();
            if (!allowed_problem_occured)
            {
                throw Fred::Registry::InternalServerError{"LibFred::CancelObjectStateRequestId::Exception caught"};
            }
        }
        if (to_set)
        {
            flags_to_set.insert(flag);
        }
    });
    if (!flags_to_set.empty())
    {
        LibFred::CreateObjectStateRequestId{info_nsset_data.id, flags_to_set}.exec(ctx);
        request_changed = true;
    }
    if (request_changed)
    {
        LibFred::PerformObjectStateRequest{info_nsset_data.id}.exec(ctx);
    }
}

}//namespace {anonymous}

void Fred::Registry::Nsset::update_nsset_state(const UpdateNssetStateRequest& request)
{
    try
    {
        LibFred::OperationContextCreator ctx;
        LibFred::OperationContextLockingForUpdate ctx_for_update{ctx};
        std::for_each(begin(request.updates), end(request.updates), [&request, &ctx_for_update](auto&& to_update)
        {
            update_state(ctx_for_update, request, to_update.first, to_update.second);
        });
        ctx.commit_transaction();
    }
    catch (const LibFred::UnknownObjectId&)
    {
        throw UpdateNssetStateReply::Exception::NssetDoesNotExist{request};
    }
    catch (const UpdateNssetStateReply::Exception::NssetDoesNotExist&)
    {
        throw;
    }
    catch (const UpdateNssetStateReply::Exception::StateFlagDoesNotExist&)
    {
        throw;
    }
    catch (const UpdateNssetStateReply::Exception::InvalidSnapshot&)
    {
        throw;
    }
    catch (const UpdateNssetStateReply::Exception::InvalidData&)
    {
        throw;
    }
    catch (const Fred::Registry::InternalServerError&)
    {
        throw;
    }
    catch (const LibFred::CreateObjectStateRequestId::Exception& e)
    {
        throw Fred::Registry::InternalServerError{"unexpected problem signalized by LibFred::CreateObjectStateRequestId::Exception"};
    }
    catch (const Database::Exception& e)
    {
        throw Fred::Registry::DatabaseError{std::string{"Database::Exception: "} + e.what()};
    }
    catch (const std::exception& e)
    {
        throw Fred::Registry::InternalServerError(e.what());
    }
    catch (...)
    {
        throw Fred::Registry::InternalServerError{"unexpected unknown exception"};
    }
}
